import json
from abc import ABCMeta, abstractproperty
from datetime import datetime, timezone
from json import JSONDecoder, JSONEncoder
from typing import Any, Set, TypeVar, Generic

from dateutil.parser import parser

from hgijson.json.interfaces import ParsedJSONDecoder
from hgijson.types import PrimitiveJsonSerializableType, SerializableType

ItemType = TypeVar("ItemType")


class StrJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__str__` to its string representation.
    """
    def default(self, to_encode: Any) -> str:
        return str(to_encode)


class StrJSONDecoder(JSONDecoder):
    """
    JSON decoder for strings.
    """
    def decode(self, to_decode: str, **kwargs) -> str:
        return str(to_decode)


class IntJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__int__`  to an integer.
    """
    def default(self, to_encode: Any) -> str:
        return int(to_encode)


class IntJSONDecoder(JSONDecoder):
    """
    JSON decoder for integers.
    """
    def decode(self, to_decode: str, **kwargs) -> int:
        return int(to_decode)


class FloatJSONEncoder(JSONEncoder):
    """
    JSON encoder from any type that implements `__float__`  to a float.
    """
    def default(self, to_encode: Any) -> str:
        return float(to_encode)


class FloatJSONDecoder(JSONDecoder):
    """
    JSON decoder for floats.
    """
    def decode(self, to_decode: str, **kwargs) -> str:
        return float(to_decode)


class DatetimeISOFormatJSONEncoder(JSONEncoder):
    """
    JSON encoder for datetime to ISO 8601 format.
    """
    def default(self, to_encode: datetime) -> str:
        return to_encode.isoformat()


class DatetimeISOFormatJSONDecoder(JSONDecoder):
    """
    JSON decoder for datetime as ISO 8601 formatted string.
    """
    _DATE_PARSER = parser()

    def decode(self, to_decode: str, **kwargs) -> datetime:
        return DatetimeISOFormatJSONDecoder._DATE_PARSER.parse(to_decode)


class DatetimeEpochJSONEncoder(JSONEncoder):
    """
    JSON encoder for datetime to seconds since the epoch (1970-01-01). If the datetime has microsecond precision, it
    will be rounded to the nearest corresponding second since the epoch.
    """
    def default(self, to_encode: datetime) -> int:
        return int(to_encode.timestamp())


class DatetimeEpochJSONDecoder(JSONDecoder):
    """
    JSON decoder for datetime as seconds since the epoch (1970-01-01).
    """
    def decode(self, to_decode: str, **kwargs) -> datetime:
        return datetime.fromtimestamp(int(to_decode), timezone.utc)


class SetJSONEncoder(Generic[ItemType], JSONEncoder, metaclass=ABCMeta):
    """
    Encoder for sets, which serialises sets into JSON lists.
    """
    @abstractproperty
    def item_encoder_cls(self) -> type:
        """
        The type of JSON encoder to use for each item in a set.
        :return: the type of item JSON encoder - must be a subclass of `JSONEncoder`
        """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._item_encoder = self.item_encoder_cls(*args, **kwargs)     # type: JSONEncoder

    def default(self, to_encode: Set[ItemType]) -> PrimitiveJsonSerializableType:
        if not isinstance(to_encode, Set):
            super().default(to_encode)
        encoded_set = []
        for item in to_encode:
            encoded_item = self._item_encoder.default(item)
            encoded_set.append(encoded_item)
        return encoded_set


class SetJSONDecoder(Generic[ItemType], JSONDecoder, ParsedJSONDecoder, metaclass=ABCMeta):
    """
    Decoder for sets, which deserialises JSON lists into Python sets.
    """
    @abstractproperty
    def item_decoder_cls(self) -> type:
        """
        The type of JSON decoder for each item in a set that has been encoded as a JSON list.
        :return: the type of item JSON decoder - must be a subclass of `JSONDecoder`
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._item_decoder = self.item_decoder_cls(*args, **kwargs)     # type: JSONDecoder

    def decode(self, to_decode: str, **kwargs) -> Set[ItemType]:
        to_decode_as_list = json.loads(to_decode)
        return self.decode_parsed(to_decode_as_list)

    def decode_parsed(self, parsed_json: PrimitiveJsonSerializableType) -> SerializableType:
        decoded_set = set()
        for item in parsed_json:
            if isinstance(self._item_decoder, ParsedJSONDecoder):
                # Optimisation: `ParsedJSONDecoder` knows how to decode a dict - no need to convert to JSON as string
                decoded_item = self._item_decoder.decode_parsed(item)
            else:
                item_as_string = json.dumps(item)
                decoded_item = self._item_decoder.decode(item_as_string)

            decoded_set.add(decoded_item)
        return decoded_set



import json
import unittest
from json import JSONDecoder, JSONEncoder

from hgicommon.models import Model

from hgijson.json._serialization import MappingJSONDecoder
from hgijson.json._serialization import MappingJSONEncoder
from hgijson.json.builders import MappingJSONEncoderClassBuilder, MappingJSONDecoderClassBuilder, \
    SetJSONEncoderClassBuilder, SetJSONDecoderClassBuilder
from hgijson.json.models import JsonPropertyMapping
from hgijson.tests._models import SimpleModel, ComplexModel
from hgijson.tests.json._helpers import create_complex_model_with_json_representation,\
    create_simple_model_with_json_representation
from hgijson.tests.json._serializers import get_simple_model_json_property_mappings, \
    get_complex_model_json_property_mappings


class _Named(Model):
    def __init__(self):
        super().__init__()
        self.name = None    # type: str


class _Identifiable(Model):
    def __init__(self):
        super().__init__()
        self.id = None    # type: int


class _Employee(_Named, _Identifiable, Model):
    def __init__(self):
        super().__init__()
        self.title = None


class TestMappingJSONEncoderClassBuilder(unittest.TestCase):
    """
    Tests for `MappingJSONEncoderClassBuilder`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_build(self):
        encoder_builder = MappingJSONEncoderClassBuilder()
        encoder_builder.target_cls = SimpleModel
        encoder_builder.mappings = get_simple_model_json_property_mappings()

        encoder_cls = encoder_builder.build()
        encoder = encoder_cls()     # type: MappingJSONEncoder

        encoded = encoder.default(self.simple_model)
        self.assertDictEqual(encoded, self.simple_model_as_json)

    def test_build_with_superclass(self):
        SimpleModelJSONEncoder = MappingJSONEncoderClassBuilder(
            SimpleModel, get_simple_model_json_property_mappings()).build()

        encoder_builder = MappingJSONEncoderClassBuilder()
        encoder_builder.target_cls = ComplexModel
        encoder_builder.mappings = get_complex_model_json_property_mappings()
        encoder_builder.superclasses = (SimpleModelJSONEncoder, )

        encoder_cls = encoder_builder.build()
        encoder = encoder_cls()     # type: MappingJSONEncoder

        encoded = encoder.default(self.complex_model)
        self.assertDictEqual(encoded, self.complex_model_as_json)

    def test_build_with_multiple_superclasses(self):
        NamedJSONEncoder = MappingJSONEncoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        IdentifiableJSONEncoder = MappingJSONEncoderClassBuilder(_Named, [
            JsonPropertyMapping("id", "id")
        ]).build()
        EmployeeJSONEncoder = MappingJSONEncoderClassBuilder(_Employee, [
           JsonPropertyMapping("title", "title"),
        ], (NamedJSONEncoder, IdentifiableJSONEncoder)).build()

        employee = _Employee()
        employee.name = "Bob"
        employee.id = 42
        employee.title = "Software Dev"
        employee_as_json = {
            "name": employee.name,
            "id": employee.id,
            "title": employee.title
        }
        self.assertEqual(EmployeeJSONEncoder().default([employee]), [employee_as_json])


class TestMappingJSONDecoderClassBuilder(unittest.TestCase):
    """
    Tests for `MappingJSONDecoderClassBuilder`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_build(self):
        decoder_builder = MappingJSONDecoderClassBuilder()
        decoder_builder.target_cls = SimpleModel
        decoder_builder.mappings = get_simple_model_json_property_mappings()

        decoder_cls = decoder_builder.build()
        decoder = decoder_cls()     # type: MappingJSONDecoder

        decoded = decoder.decode(json.dumps(self.simple_model_as_json))
        self.assertEqual(decoded, self.simple_model)

    def test_build_with_superclass(self):
        SimpleModelJSONDecoder = MappingJSONDecoderClassBuilder(
            target_cls=SimpleModel, mappings=get_simple_model_json_property_mappings()).build()

        decoder_builder = MappingJSONDecoderClassBuilder()
        decoder_builder.superclasses = (SimpleModelJSONDecoder, )
        decoder_builder.target_cls = ComplexModel
        decoder_builder.mappings = get_complex_model_json_property_mappings()

        decoder_cls = decoder_builder.build()
        decoder = decoder_cls()     # type: MappingJSONDecoder

        decoded = decoder.decode(json.dumps(self.complex_model_as_json))
        self.assertEqual(decoded, self.complex_model)

    def test_build_with_multiple_superclasses(self):
        NamedJSONDecoder = MappingJSONDecoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        IdentifiableJSONDecoder = MappingJSONDecoderClassBuilder(_Named, [
            JsonPropertyMapping("id", "id")
        ]).build()
        EmployeeJSONDecoder = MappingJSONDecoderClassBuilder(_Employee, [
           JsonPropertyMapping("title", "title"),
        ], (NamedJSONDecoder, IdentifiableJSONDecoder)).build()

        employee = _Employee()
        employee.name = "Bob"
        employee.id = 42
        employee.title = "Software Dev"
        employee_as_json = {
            "name": employee.name,
            "id": employee.id,
            "title": employee.title
        }
        employee_as_json_string = json.dumps([employee_as_json])
        self.assertEqual(EmployeeJSONDecoder().decode(employee_as_json_string), [employee])


class TestSetJSONEncoderClassBuilder(unittest.TestCase):
    """
    Tests for `SetJSONEncoderClassBuilder`.
    """
    def setUp(self):
        self.values = set([create_simple_model_with_json_representation(i)[0] for i in range(10)])
        self.values_as_json = [create_simple_model_with_json_representation(i)[1] for i in range(10)]
        self.ValueJSONEncoder = MappingJSONEncoderClassBuilder(
            SimpleModel, get_simple_model_json_property_mappings()).build()

    def test_build(self):
        CustomSetJSONEncoder = SetJSONEncoderClassBuilder(self.ValueJSONEncoder).build()
        encoder = CustomSetJSONEncoder()    # type: JSONEncoder
        self.assertCountEqual(encoder.default(self.values), self.values_as_json)


class TestSetJSONDecoderClassBuilder(unittest.TestCase):
    """
    Tests for `SetJSONDecoderClassBuilder`.
    """
    def setUp(self):
        self.values = set([create_simple_model_with_json_representation(i)[0] for i in range(10)])
        self.values_as_json = [create_simple_model_with_json_representation(i)[1] for i in range(10)]
        self.ValueJSONDecoder = MappingJSONDecoderClassBuilder(
            SimpleModel, get_simple_model_json_property_mappings()).build()

    def test_build(self):
        CustomSetJSONDecoder = SetJSONDecoderClassBuilder(self.ValueJSONDecoder).build()
        decoder = CustomSetJSONDecoder()    # type: JSONDecoder
        values_as_json_string = json.dumps(self.values_as_json)
        self.assertEqual(decoder.decode(values_as_json_string), self.values)


if __name__ == "__main__":
    unittest.main()



from json import JSONEncoder

from hgicommon.models import Model
from hgijson.json.automatic import _RegisteredTypeJSONEncoder


class StubModel(Model):
    """
    Stub `Model`.
    """
    pass


class StubRegisteredTypeJSONEncoder(_RegisteredTypeJSONEncoder):
    """
    Stub `_RegisteredTypeJSONEncoder`.
    """
    def _get_json_encoders_for_type(self) -> JSONEncoder:
        pass



from hgijson.json.primitive import SetJSONEncoder, SetJSONDecoder


class MockSetJSONEncoder(SetJSONEncoder):
    """
    Mock implementation of `SetJSONEncoder`.
    """
    def __init__(self, item_encoder_cls: type, *args, **kwargs):
        self._injected_item_encoder_cls = item_encoder_cls
        super().__init__(*args, **kwargs)

    @property
    def item_encoder_cls(self) -> type:
        return self._injected_item_encoder_cls


class MockSetJSONDecoder(SetJSONDecoder):
    """
    Mock implementation of `SetJSONDecoder`.
    """
    def __init__(self, item_decoder_cls: type, *args, **kwargs):
        self._injected_item_decoder_cls = item_decoder_cls
        super().__init__(*args, **kwargs)

    @property
    def item_decoder_cls(self) -> type:
        return self._injected_item_decoder_cls



from json import JSONDecoder, JSONEncoder
from typing import Callable, Any, Dict

from hgijson.json._converters import json_decoder_to_deserializer, json_encoder_to_serializer
from hgijson.models import PropertyMapping


class JsonPropertyMapping(PropertyMapping):
    """
    Model of a mapping between a json property and a property of an object.
    """
    def __init__(
            self, json_property_name=None, object_property_name: str=None, object_constructor_parameter_name: str=None,
            json_property_getter: Callable[[Dict], Any]=None, json_property_setter: Callable[[Any, Any], None]=None,
            encoder_cls: type=JSONEncoder, decoder_cls: type=JSONDecoder, optional: bool=False, **kwargs):
        """
        Constructor.
        :param json_property_name:
        :param object_property_name:
        :param object_constructor_parameter_name:
        :param json_property_getter:
        :param json_property_setter:
        :param encoder_cls:
        :param decoder_cls:
        :param optional:
        """
        if json_property_name is not None:
            if json_property_getter is not None and json_property_setter is not None:
                raise ValueError("Redundant `json_property_name` argument given. It has been specified that a "
                                 "serialized property is to be used via the given property name and both a setter and "
                                 "getter of this property has been provided. To avoid confusion, the serialized "
                                 "property cannot be specified in this case.")

            if json_property_getter is None:
                def json_property_getter(obj_as_json: dict):
                    if optional and json_property_name not in obj_as_json:
                        return None
                    return obj_as_json[json_property_name]

            if json_property_setter is None:
                def json_property_setter(obj_as_json: dict, value: Any):
                    obj_as_json[json_property_name] = value

        encoder_as_serializer_cls = json_encoder_to_serializer(encoder_cls)
        decoder_as_serializer_cls = json_decoder_to_deserializer(decoder_cls)

        super().__init__(object_property_name, object_constructor_parameter_name,
                         serialized_property_getter=json_property_getter,
                         serialized_property_setter=json_property_setter,
                         serializer_cls=encoder_as_serializer_cls, deserializer_cls=decoder_as_serializer_cls,
                         optional=optional, **kwargs)

    @property
    def json_property_getter(self) -> Callable[[Dict], Any]:
        return self.serialized_property_getter

    @json_property_getter.setter
    def json_property_getter(self, getter: Callable[[Dict], Any]):
        self.serialized_property_getter = getter

    @property
    def json_property_setter(self) -> Callable[[Any, Any], None]:
        return self.serialized_property_setter

    @json_property_setter.setter
    def json_property_setter(self, setter: Callable[[Any, Any], None]):
        self.serialized_property_setter = setter



from hgijson.json.primitive import IntJSONEncoder, StrJSONEncoder, IntJSONDecoder, FloatJSONEncoder, FloatJSONDecoder, \
    DatetimeISOFormatJSONEncoder, DatetimeISOFormatJSONDecoder, DatetimeEpochJSONEncoder, DatetimeEpochJSONDecoder, \
    SetJSONEncoder, SetJSONDecoder, ItemType, StrJSONDecoder

from hgijson.json.builders import MappingJSONDecoderClassBuilder, SetJSONEncoderClassBuilder, \
    SetJSONDecoderClassBuilder, MappingJSONEncoderClassBuilder

from hgijson.json.models import JsonPropertyMapping





from abc import ABCMeta, abstractmethod
from typing import Any, Generic, Iterable, Union, Sequence, Dict, List

from hgijson.types import SerializableType, PrimitiveUnionType, PrimitiveJsonSerializableType


class Serializer(Generic[SerializableType, PrimitiveUnionType], metaclass=ABCMeta):
    """
    Serializer that uses a mapping that describes how the process should occur for the type of serializable object that
    this serializer handles.
    """
    # TODO: Correct type hinting in signature without causing a cyclic dependency issue
    @abstractmethod
    # def _create_serializer_of_type(self, serializer_type: type) -> Serializer:
    def _create_serializer_of_type(self, serializer_type: type):
        """
        Create an instance of an serializer of the given type.
        :param serializer_type: the type of serializer to instantiate (a subclass of `Serializer`)
        :return: the created serializer
        """

    @abstractmethod
    def _create_serialized_container(self) -> Any:
        """
        Create the container in which serialized representation is built in
        :return: the container
        """

    def __init__(self, property_mappings: Iterable):
        """
        Constructor.
        :param property_mappings: TODO
        """
        self._property_mappings = property_mappings     # type_but_do_not_import: Iterable[PropertyMapping]
        self._serializers_cache = dict()    # type: Dict[type, Serializer]

    def serialize(self, serializable: Union[SerializableType, Sequence[SerializableType]]) -> PrimitiveUnionType:
        """
        Serializes the given serializable object or collection of serializable objects.
        :param serializable: the object to serialize
        :return: a serialization of the given object
        """
        if isinstance(serializable, List):
            return [self.serialize(item) for item in serializable]
        else:
            serialized = self._create_serialized_container()

            for mapping in self._property_mappings:
                if mapping.object_property_getter is not None and mapping.serialized_property_setter is not None:
                    value = mapping.object_property_getter(serializable)
                    if not (mapping.optional and value is None):
                        encoded_value = self._serialize_property_value(value, mapping.serializer_cls)
                        mapping.serialized_property_setter(serialized, encoded_value)

            return serialized

    def _serialize_property_value(self, to_serialize: Any, serializer_cls: type) -> Any:
        """
        Serializes the given value using the given serializer.
        :param to_serialize: the value to deserialize
        :param serializer_cls: the type of serializer to use
        :return: serialized value
        """
        serializer = self._create_serializer_of_type_with_cache(serializer_cls)
        assert serializer is not None
        return serializer.serialize(to_serialize)

    # TODO: Fix self-referential signature
    def _create_serializer_of_type_with_cache(self, serializer_type: type):
        """
        Creates a deserializer of the given type, exploiting a cache.
        :param serializer_type: the type of deserializer to create
        :return: the created serializer
        """
        if serializer_type not in self._serializers_cache:
            self._serializers_cache[serializer_type] = self._create_serializer_of_type(serializer_type)
        return self._serializers_cache[serializer_type]


class Deserializer(Generic[SerializableType, PrimitiveUnionType], metaclass=ABCMeta):
    """
    Deserializer that uses a mapping that describes how the process should occur for the type of deserializable object
    that this deserializer handles.
    """
    @abstractmethod
    def _create_deserializer_of_type(self, deserializer_type: type):
        """
        Creates a deserializer of the given type.
        :param deserializer_type: the type of deserializer to create
        :return: the created deserializer (of type `Deserializer`)
        """

    def __init__(self, property_mappings: Iterable, deserializable_cls: type):
        """
        Constructor.
        :param property_mappings: TODO
        :param deserializable_cls: the class that should be built as a result of deserialization
        """
        self._property_mappings = property_mappings     # type_but_do_not_import: Iterable[PropertyMapping]
        self._deserializable_cls = deserializable_cls
        self._deserializers_cache = dict()    # type: Dict[type, Deserializer]

    def deserialize(self, to_deserialize: PrimitiveJsonSerializableType) \
            -> Union[SerializableType, Sequence[SerializableType]]:
        """
        Deserializes the given representation of the serialized object.
        :param to_deserialize: the serialized object as a dictionary
        :return: the deserialized object or collection of deserialized objects
        """
        if isinstance(to_deserialize, list):
            deserialized = []
            for item in to_deserialize:
                item_deserialized = self.deserialize(item)
                deserialized.append(item_deserialized)
            return deserialized
        else:
            mappings_not_set_in_constructor = []    # type_but_do_not_import: List[PropertyMapping]

            init_kwargs = dict()    # type: Dict[str, Any]
            for mapping in self._property_mappings:
                if mapping.object_constructor_parameter_name is not None:
                    value = mapping.serialized_property_getter(to_deserialize)
                    if not (mapping.optional and value is None):
                        decoded_value = self._deserialize_property_value(value, mapping.deserializer_cls)
                        argument = mapping.object_constructor_argument_modifier(decoded_value)
                        init_kwargs[mapping.object_constructor_parameter_name] = argument
                else:
                    mappings_not_set_in_constructor.append(mapping)

            decoded = self._deserializable_cls(**init_kwargs)
            assert type(decoded) == self._deserializable_cls

            for mapping in mappings_not_set_in_constructor:
                assert mapping.object_constructor_parameter_name is None
                if mapping.serialized_property_getter is not None and mapping.object_property_setter is not None:
                    value = mapping.serialized_property_getter(to_deserialize)
                    if not (mapping.optional and value is None):
                        decoded_value = self._deserialize_property_value(value, mapping.deserializer_cls)
                        mapping.object_property_setter(decoded, decoded_value)

            return decoded

    def _deserialize_property_value(self, to_deserialize: PrimitiveJsonSerializableType, deserializer_cls: type) -> Any:
        """
        Deserializes the given value using the given deserializer.
        :param to_deserialize: the value to deserialize
        :param deserializer_cls: the type of deserializer to use
        :return: deserialized value
        """
        deserializer = self._create_deserializer_of_type_with_cache(deserializer_cls)
        assert deserializer is not None
        return deserializer.deserialize(to_deserialize)

    # TODO: Fix self-referential signature
    def _create_deserializer_of_type_with_cache(self, deserializer_type: type):
        """
        Creates a deserializer of the given type, exploiting a cache.
        :param deserializer_type: the type of deserializer to create
        :return: the deserializer
        """
        if deserializer_type not in self._deserializers_cache:
            self._deserializers_cache[deserializer_type] = self._create_deserializer_of_type(deserializer_type)
        return self._deserializers_cache[deserializer_type]



from typing import Dict, Callable, Any, Set

from hgijson.serializers import PrimitiveDeserializer, PrimitiveSerializer


class PropertyMapping():
    """
    Model of a mapping between a json property and a property of an object.
    """
    def __init__(
            self, object_property_name: str=None, object_constructor_parameter_name: str=None,
            object_constructor_argument_modifier: Callable[[Any], Any]=None,
            serialized_property_getter: Callable[[Dict], Any]=None,
            serialized_property_setter: Callable[[Any, Any], None]=None,
            object_property_getter: Callable[[Any], Any]=None, object_property_setter: Callable[[Any, Any], None]=None,
            serializer_cls: type=PrimitiveSerializer, deserializer_cls: type=PrimitiveDeserializer,
            optional: bool=False):
        """
        Constructor.
        :param object_property_name: defines the object property to assign the value returned by
        `serialized_property_getter` to using a simple `object_property_getter` function
        :param object_constructor_parameter_name: defines the constructor parameter to bind the value returned by
        `serialized_property_getter` to using a simple `serialized_property_getter` function
        :param object_constructor_argument_modifier: constructor argument modification function
        :param serialized_property_getter: method that, given the serialized object, gets the property of interest to
        this mapping
        :param serialized_property_setter: method that, given the serialized object, sets a serialized property to the
        value returned by `object_property_getter`
        :param object_property_getter: method that, given the object, gets the property of interest to this mapping
        :param object_property_setter: method that, given the object, sets a property to the value returned by
        `serialized_property_getter`
        :param serializer_cls: class to serialize the value returned by `object_property_getter`
        :param deserializer_cls: class to deserialize the value returned by `serialized_property_getter`
        :param optional: whether the property is optional - will ignore if `None` in serialized representation and will
        not serialize if `None` in object
        """
        if object_property_name is not None:
            # It is not possible to know what the serialized container is therefore default values cannot be set if a
            # required getter/setter is not given. Defaults can however be set in subclasses.

            if object_property_getter is not None and object_property_setter is not None:
                raise ValueError("Redundant `object_property_name` argument given. It has been specified that an "
                                 "object property is to be used via the given property name and both a setter and "
                                 "getter of this property has been provided. To avoid confusion, the object "
                                 "property cannot be specified in this case.")

            if object_property_getter is None:
                def object_property_getter(obj: Any) -> Any:
                    return obj.__getattribute__(object_property_name)

            if object_property_setter is None and object_constructor_parameter_name is None:
                def object_property_setter(obj: Any, value: Any):
                    if not hasattr(obj, object_property_name):
                        raise AttributeError("Object `%s` does not have the attribute `%s`"
                                             % (obj, object_property_name))
                    obj.__setattr__(object_property_name, value)

        if object_constructor_parameter_name is not None:
            if serialized_property_getter is None:
                raise ValueError("`serialized_property_getter` must be defined alongside "
                                 "`object_constructor_parameter_name`.")

            if object_property_setter is not None and object_constructor_parameter_name is not None:
                raise ValueError("`object_property_setter` cannot be defined if `object_constructor_parameter_name` is "
                                 "given as the serialized data will be injected into the object as an argument. If an "
                                 "object property should also be set using a setter, define another mapping. If the "
                                 "constructor argument needs to be modified, use "
                                 "`object_constructor_argument_modifier`.")

            if object_constructor_argument_modifier is None:
                def object_constructor_argument_modifier(argument: Any) -> Any:
                    return argument
        else:
            if object_constructor_argument_modifier is not None:
                raise ValueError("`object_constructor_argument_modifier` cannot be used without "
                                 "`object_constructor_parameter_name` being set.")

        self.serialized_property_getter = serialized_property_getter
        self.serialized_property_setter = serialized_property_setter
        self.object_constructor_parameter_name = object_constructor_parameter_name
        self.object_constructor_argument_modifier = object_constructor_argument_modifier
        self.object_property_getter = object_property_getter
        self.object_property_setter = object_property_setter
        self.serializer_cls = serializer_cls
        self.deserializer_cls = deserializer_cls
        self.optional = optional

    def __str__(self) -> str:
        string_builder = []
        for property, value in vars(self).items():
            string_builder.append("%s: %s" % (property, value))
        string_builder = sorted(string_builder)
        return "{ %s }" % ', '.join(string_builder)

    def __repr__(self) -> str:
        return "<%s object at %s: %s>" % (type(self), id(self), str(self))



from typing import List

from hgicommon.models import Model


class SimpleModel(Model):
    """
    Example of a simple model.
    """
    def __init__(self, constructor_b=None):
        self.a = None
        self.b = constructor_b


class ComplexModel(SimpleModel):
    """
    Example of a complex model.
    """
    def __init__(self, constructor_b):
        super().__init__(constructor_b)
        self.c = None
        self.d = [SimpleModel(constructor_b + i) for i in range(3)]
        self.e = [4.1, 5.2, 6.3]
        self.f = {"1": 2, "3": 4}
        self.g = "test"
        self.h = 123

        for i in range(len(self.d)):
            self.d[i].a = i

    def all(self) -> List:
        return [self.a, self.b, self.c, self.d]



import unittest
from typing import Dict

from hgijson.json.models import JsonPropertyMapping
from hgijson.models import PropertyMapping
from hgijson.tests._models import SimpleModel
from hgijson.tests._serializers import ComplexModelSerializer, SimpleModelDeserializer, ComplexModelDeserializer, \
    SimpleModelSerializer
from hgijson.tests.json._helpers import create_complex_model_with_json_representation, \
    create_simple_model_with_json_representation


class TestSerializer(unittest.TestCase):
    """
    Tests for `Serializer`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_serialize_with_no_mapping(self):
        mappings = [PropertyMapping()]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.simple_model), {})

    def test_serialize_using_names(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"), JsonPropertyMapping("serialized_b", "b")]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.simple_model), self.simple_model_as_json)

    def test_serialize_using_object_property_getter(self):
        mappings = [JsonPropertyMapping("serialized_b_plus_one", object_property_getter=lambda model: model.b + 1)]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.simple_model), {"serialized_b_plus_one": self.simple_model.b + 1})

    def test_serialize_using_serialized_property_setter(self):
        def serialized_property_setter(container: Dict, value: Dict):
            container["augmented_key_%s" % value] = value

        mappings = [JsonPropertyMapping(None, "b", json_property_setter=serialized_property_setter)]
        serializer = ComplexModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.complex_model), {"augmented_key_5": 5})

    def test_serialize_using_serializer_cls(self):
        def serialized_property_setter_for_d(container: Dict, value: Dict):
            container["serialized_d"] = value

        mappings = [PropertyMapping("d", serialized_property_setter=serialized_property_setter_for_d,
                                    serializer_cls=SimpleModelSerializer)]
        serializer = ComplexModelSerializer(mappings)

        complex_model = self.complex_model
        expected = {"serialized_d": [{
            "serialized_a": i,
            "serialized_b": complex_model.b + i
        } for i in range(len(complex_model.d))]}
        self.assertDictEqual(serializer.serialize(complex_model), expected)

    def test_serialize_with_iterable_with_no_items(self):
        serialized = SimpleModelSerializer().serialize([])
        self.assertCountEqual(serialized, [])

    def test_serialize_with_iterable(self):
        complex_models = [create_complex_model_with_json_representation(i)[0] for i in range(10)]
        complex_models_as_json = [create_complex_model_with_json_representation(i)[1] for i in range(10)]
        serialized = ComplexModelSerializer().serialize(complex_models)
        self.assertEqual(serialized, complex_models_as_json)

    def test_serialize_optional_when_not_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(SimpleModel()), {})

    def test_serialize_optional_when_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        serializer = SimpleModelSerializer(mappings)
        self.assertDictEqual(serializer.serialize(self.simple_model), self.simple_model_as_json)

    def test_serialize_when_property_does_not_exist(self):
        mappings = [JsonPropertyMapping("serialized_a", "z")]
        serializer = SimpleModelSerializer(mappings)
        self.assertRaises(AttributeError, serializer.serialize, self.simple_model)


class TestDeserializer(unittest.TestCase):
    """
    Tests for `Deserializer`.
    """
    def setUp(self):
        self.simple_model = create_simple_model_with_json_representation()[0]
        self.simple_model_as_json = create_simple_model_with_json_representation()[1]
        self.complex_model = create_complex_model_with_json_representation()[0]
        self.complex_model_as_json = create_complex_model_with_json_representation()[1]

    def test_deserialize_with_no_mapping(self):
        mappings = [PropertyMapping()]
        deserializer = SimpleModelDeserializer(mappings)
        self.simple_model.a = None
        self.simple_model.b = None
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), self.simple_model)

    def test_deserialize_using_names(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"), JsonPropertyMapping("serialized_b", "b")]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), self.simple_model)

    def test_deserialize_using_constructor_parameter(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"),
                    JsonPropertyMapping("serialized_b", object_constructor_parameter_name="constructor_b")]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), self.simple_model)

    def test_deserialize_using_constructor_parameter_with_modifier(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", object_constructor_parameter_name="constructor_b",
                                        object_constructor_argument_modifier=lambda value: value + 1)]
        deserializer = SimpleModelDeserializer(mappings)
        model = SimpleModel(self.simple_model_as_json["serialized_a"] + 1)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), model)

    def test_deserialize_using_deserializer(self):
        mappings = [JsonPropertyMapping("serialized_a", "a"),
                    JsonPropertyMapping("serialized_b", object_constructor_parameter_name="constructor_b"),
                    JsonPropertyMapping("serialized_c", "c"),
                    PropertyMapping("d", serialized_property_getter=lambda obj_as_dict: obj_as_dict["serialized_d"],
                                    deserializer_cls=SimpleModelDeserializer)]
        deserializer = ComplexModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(self.complex_model_as_json), self.complex_model)

    def test_deserialize_using_serialized_property_getter(self):
        a_and_b = {
            "ab": [self.simple_model.a, self.simple_model.b]
        }
        mappings = [JsonPropertyMapping(None, "a", json_property_getter=lambda json_as_dict: json_as_dict["ab"][0]),
                    JsonPropertyMapping(None, "b", json_property_getter=lambda json_as_dict: json_as_dict["ab"][1])]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(a_and_b), self.simple_model)

    def test_deserialize_with_iterable_with_no_items(self):
        decoded = SimpleModelDeserializer().deserialize([])
        self.assertEqual(decoded, [])

    def test_deserialize_with_iterable(self):
        complex_models = [create_complex_model_with_json_representation(i)[0] for i in range(10)]
        complex_models_as_json = [create_complex_model_with_json_representation(i)[1] for i in range(10)]
        decoded = ComplexModelDeserializer().deserialize(complex_models_as_json)
        self.assertEqual(decoded, complex_models)

    def test_deserialize_optional_when_not_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize({}), SimpleModel())

    def test_deserialize_optional_when_set(self):
        mappings = [JsonPropertyMapping("serialized_a", "a", optional=True),
                    JsonPropertyMapping("serialized_b", "b", optional=True)]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertEqual(deserializer.deserialize(self.simple_model_as_json), self.simple_model)

    def test_deserialize_when_property_does_not_exist(self):
        mappings = [JsonPropertyMapping("serialized_a", "z")]
        deserializer = SimpleModelDeserializer(mappings)
        self.assertRaises(AttributeError, deserializer.deserialize, self.simple_model_as_json)


if __name__ == "__main__":
    unittest.main()



import json
from abc import ABCMeta, abstractproperty
from json import JSONDecoder, JSONEncoder
from typing import Any, Dict

from hgijson.json.interfaces import ParsedJSONDecoder
from hgijson.serialization import Deserializer, Serializer
from hgijson.types import PrimitiveJsonSerializableType, PrimitiveUnionType, SerializableType

_serializer_cache = dict()  # type: Dict[type, Serializer]
_deserializer_cache = dict()    # type: Dict[type, Deserializer]

_PLACEHOLDER = object


class _JSONEncoderAsSerializer(Serializer, metaclass=ABCMeta):
    """
    JSON encoder wrapped to work as a `Serializer`.
    """
    @abstractproperty
    def _get_encoder_type(self, *args) -> type:
        """
        Gets the `JSONEncoder` type that is used by this serialiser.
        :return: the encoder type (must be a subclass of `JSONEncoder`)
        """

    def __init__(self, *args, **kwargs):
        super().__init__(())
        json_encoder_cls = self._get_encoder_type()
        self._encoder = json_encoder_cls(*args, **kwargs)  # type: JSONEncoder

    def serialize(self, serializable: SerializableType) -> PrimitiveUnionType:
        if type(self._encoder) == JSONEncoder:
            # Have to load from string to more rich representation - not possible to stop `encode` going to string :/
            return json.loads(self._encoder.encode(serializable))
        else:
            return self._encoder.default(serializable)

    def _create_serializer_of_type(self, serializer_type: type):
        """
        Unused - implemented to satisfy the interface only.
        """

    def _create_serialized_container(self) -> Any:
        """
        Unused - implemented to satisfy the interface only.
        """


class _JSONDecoderAsDeserializer(Deserializer, metaclass=ABCMeta):
    """
    JSON decoder wrapped to work as a `Deserializer`.
    """
    @abstractproperty
    def _get_decoder_type(self, *args) -> type:
        """
        Gets the `JSONDecoder` type that is used by this deserialiser.
        :return: the decoder type (must be a subclass of `JSONDecoder`)
        """

    def __init__(self, *args, **kwargs):
        json_decoder_cls = self._get_decoder_type()
        super().__init__((), json_decoder_cls)
        self._decoder = json_decoder_cls(*args, **kwargs)  # type: JSONDecoder

    def deserialize(self, json_as_dict: PrimitiveJsonSerializableType) -> SerializableType:
        if not isinstance(self._decoder, ParsedJSONDecoder):
            # Decode must take a string (even though we have a richer representation) :/
            json_as_string = json.dumps(json_as_dict)
            return self._decoder.decode(json_as_string)
        else:
            # Optimisation - no need to convert our relatively rich representation into a string (just to turn it back
            # again!)
            return self._decoder.decode_parsed(json_as_dict)

    def _create_deserializer_of_type(self, deserializer_type: type):
        """
        Unused - implemented to satisfy the interface only.
        """


def json_encoder_to_serializer(encoder_cls: type) -> type:
    """
    Converts a `JSONEncoder` class into an equivalent `Serializer` class.
    :param encoder_cls: the encoder class
    :return: the equivalent `Serializer` class
    """
    if encoder_cls not in _serializer_cache:
        encoder_as_serializer_cls = type(
            "%sAsSerializer" % encoder_cls.__class__.__name__,
            (_JSONEncoderAsSerializer,),
            {
                "_get_encoder_type": lambda self: encoder_cls
            }
        )
        _serializer_cache[encoder_cls] = encoder_as_serializer_cls

    return _serializer_cache[encoder_cls]


def json_decoder_to_deserializer(decoder_cls: type) -> type:
    """
    Converts a `JSONDecoder` class into an equivalent `Deserializer` class.
    :param decoder_cls: the decoder class
    :return: the equivalent `Deserializer` class
    """
    if decoder_cls not in _deserializer_cache:
        encoder_as_deserializer_cls = type(
            "%sAsDeserializer" % decoder_cls.__class__.__name__,
            (_JSONDecoderAsDeserializer,),
            {
                "_get_decoder_type": lambda self: decoder_cls
            }
        )
        _deserializer_cache[decoder_cls] = encoder_as_deserializer_cls

    return _deserializer_cache[decoder_cls]



import json
import unittest
from datetime import datetime, timezone

from hgijson.json.primitive import StrJSONDecoder, IntJSONEncoder, FloatJSONEncoder, FloatJSONDecoder, \
    DatetimeEpochJSONEncoder, DatetimeEpochJSONDecoder, DatetimeISOFormatJSONDecoder, DatetimeISOFormatJSONEncoder, \
    IntJSONDecoder
from hgijson.json.primitive import StrJSONEncoder
from hgijson.tests.json._mocks import MockSetJSONEncoder, MockSetJSONDecoder


class TestStrJSONEncoder(unittest.TestCase):
    """
    Tests for `StrJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual(StrJSONEncoder().default("123"), "123")

    def test_default_with_int(self):
        self.assertEqual(StrJSONEncoder().default(123), "123")

    def test_default_with_float(self):
        self.assertEqual(StrJSONEncoder().default(12.3), "12.3")


class TestStrJSONDecoder(unittest.TestCase):
    """
    Tests for `TestStrJSONDecoder`.
    """
    def test_decode_with_string(self):
        self.assertEqual(StrJSONDecoder().decode("123"), "123")

    def test_with_json_loads(self):
        self.assertEqual(json.loads("123", cls=StrJSONDecoder), "123")


class TestIntJSONEncoder(unittest.TestCase):
    """
    Tests for `IntJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual(IntJSONEncoder().default("123"), 123)

    def test_default_with_int(self):
        self.assertEqual(IntJSONEncoder().default(123), 123)

    def test_default_with_float(self):
        self.assertEqual(IntJSONEncoder().default(12.3), 12)


class TestIntJSONDecoder(unittest.TestCase):
    """
    Tests for `IntJSONDecoder`.
    """
    def test_decode_with_int_as_string(self):
        self.assertEqual(IntJSONDecoder().decode("123"), 123)

    def test_with_json_loads_and_int_as_string(self):
        self.assertEqual(json.loads("123", cls=IntJSONDecoder), 123)


class TestFloatJSONEncoder(unittest.TestCase):
    """
    Tests for `FloatJSONEncoder`.
    """
    def test_default_with_str(self):
        self.assertEqual(FloatJSONEncoder().default("12.3"), 12.3)

    def test_default_with_int(self):
        self.assertEqual(FloatJSONEncoder().default(123), 123.0)

    def test_default_with_float(self):
        self.assertEqual(FloatJSONEncoder().default(12.3), 12.3)


class TestFloatJSONDecoder(unittest.TestCase):
    """
    Tests for `FloatJSONDecoder`.
    """
    def test_decode_with_float_as_string(self):
        self.assertEqual(FloatJSONDecoder().decode("12.3"), 12.3)

    def test_with_json_loads_and_int_as_string(self):
        self.assertEqual(json.loads("12.3", cls=FloatJSONDecoder), 12.3)


class TestDatetimeEpochJSONEncoder(unittest.TestCase):
    """
    Tests for `DatetimeEpochJSONEncoder`.
    """
    def test_default(self):
        value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(DatetimeEpochJSONEncoder().default(value), 0)


class TestDatetimeEpochJSONDecoder(unittest.TestCase):
    """
    Tests for `DatetimeEpochJSONDecoder`.
    """
    def test_decode(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(DatetimeEpochJSONDecoder().decode("0"), expected_value)

    def test_with_json_loads(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(json.loads("0", cls=DatetimeEpochJSONDecoder), expected_value)


class TestDatetimeISOFormatJSONEncoder(unittest.TestCase):
    """
    Tests for `DatetimeISOFormatJSONEncoder`.
    """
    def test_default(self):
        value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(DatetimeISOFormatJSONEncoder().default(value), "1970-01-01T00:00:00+00:00")


class TestDatetimeISOFormatJSONDecoder(unittest.TestCase):
    """
    Tests for `DatetimeISOFormatJSONDecoder`.
    """
    def test_decode_with_numerical_offset(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(DatetimeISOFormatJSONDecoder().decode("1970-01-01T00:00:00+00:00"), expected_value)

    def test_decode_with_special_representation_offset(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(DatetimeISOFormatJSONDecoder().decode("1970-01-01T00:00:00Z"), expected_value)

    def test_with_json_loads(self):
        expected_value = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.assertEqual(json.loads("1970-01-01T00:00:00+00:00", cls=DatetimeISOFormatJSONDecoder), expected_value)


class TestSetJSONEncoder(unittest.TestCase):
    """
    Tests for `SetJSONEncoder`.
    """
    def setUp(self):
        self.item_encoder_cls = IntJSONEncoder
        self.encoder = MockSetJSONEncoder(self.item_encoder_cls)
        self.values = {1, 5, 7, 9}

    def test_default(self):
        self.assertEqual(self.encoder.encode(self.values), json.dumps(list(self.values)))

    def test_default_with_unsupported_type(self):
        self.assertRaises(TypeError, self.encoder.encode, object())


class TestSetJSONDecoder(unittest.TestCase):
    """
    Tests for `SetJSONDecoder`.
    """
    def setUp(self):
        self.item_decoder_cls = IntJSONDecoder
        self.decoder = MockSetJSONDecoder(self.item_decoder_cls)
        self.values = {1, 5, 7, 9}

    def test_decode(self):
        values_as_json_string = json.dumps(list(self.values))
        self.assertEqual(self.decoder.decode(values_as_json_string), self.values)


if __name__ == "__main__":
    unittest.main()



from typing import TypeVar, Dict, List, Tuple

SerializableType = TypeVar("Serializable")

PrimitiveUnionType = TypeVar("PrimitiveUnion")

PrimitiveJsonSerializableType = TypeVar("PrimitiveJsonSerializable", Dict, List, Tuple, str, int, float, bool, None)